from typing import Dict, List, Tuple

import numpy as np

from sparks.models import StateType, SimulationData
from sparks.simulators import DeterministicSimulator, StochasticSimulator
from ..model import (
    Model,
    ModelState,
    RateCoefficients,
    SequenceModel,
    SequenceModelState,
)


def k_CRP2_from_ratios(
    rA: float,
    rB: float,
    rX: float = 1.0,
    KAA: float = 0.0,
    KAB: float = 0.0,
    KBA: float = 0.0,
    KBB: float = 0.0,
    kpAA: float = 1.0,
) -> RateCoefficients:

    # rX = kpAA / kpBB
    kpBB = kpAA / rX

    kpAB = kpAA / rA
    kpBA = kpBB / rB

    kdAA = KAA * kpAA
    kdAB = KAB * kpAB
    kdBA = KBA * kpBA
    kdBB = KBB * kpBB

    return RateCoefficients(
        kpAA=kpAA,
        kpAB=kpAB,
        kpBA=kpBA,
        kpBB=kpBB,
        kdAA=kdAA,
        kdAB=kdAB,
        kdBA=kdBA,
        kdBB=kdBB,
    )


def k_CRP2_equal_from_ratio(
    rA: float,
    rB: float,
    K: float,
    **kwargs,
) -> RateCoefficients:

    return k_CRP2_from_ratios(rA=rA, rB=rB, KAA=K, KAB=K, KBA=K, KBB=K, **kwargs)


def c0_CRP2(
    fA0: float = 0.50, M0: float = 1.0, M0_I0_ratio: float = 1000.0
) -> ModelState:

    A0 = fA0 * M0
    B0 = (1 - fA0) * M0
    R0 = M0 / M0_I0_ratio

    return ModelState(R=R0, A=A0, B=B0)


def get_ode_model(
    T_C: float,
    fA0: float,
    M0: float,
    M2Iratio: float = 1000,
    rA: float = 15.0,
    rB: float = 0.5,
    KAA: float = 0.0,
    **kwargs,
) -> Tuple[SimulationData[ModelState], SimulationData[SequenceModelState]]:

    A0, B0 = fA0 * M0, (1 - fA0) * M0
    k = k_CRP2_from_ratios(rA, rB, KAA=KAA)
    init_state = ModelState(
        R=M0 / M2Iratio,
        A=A0,
        B=B0,
    )
    print(f"rA: {rA}, rB: {rB}, KAA: {KAA}")

    t_eval = np.linspace(0, 5e4, 10000)
    t_eval = kwargs.pop("t_eval", t_eval)

    ds = DeterministicSimulator(Model(), k)
    data = ds.simulate(
        t_eval=t_eval,
        init_state=init_state,
    )

    ds_seq = DeterministicSimulator(SequenceModel(), k)
    seq_data = ds_seq.simulate(
        t_eval=t_eval, init_state=SequenceModelState(), s=ds.ode_sol
    )

    return data, seq_data


def get_kmc_model(
    T_C: float,
    fA0: float,
    M0: float,
    M2Iratio: float = 1000,
    rA: float = 15.0,
    rB: float = 0.5,
    KAA: float = 0.0,
    **kwargs,
) -> Tuple[
    SimulationData[ModelState],
    Dict[str, SimulationData[StateType] | List[SimulationData[StateType]]],
]:

    A0, B0 = fA0 * M0, (1 - fA0) * M0
    k = k_CRP2_from_ratios(rA, rB, KAA=KAA)
    init_state = ModelState(
        R=M0 / M2Iratio,
        A=A0,
        B=B0,
    )
    print(f"rA: {rA}, rB: {rB}, KAA: {KAA}")

    ss = StochasticSimulator(Model(), k, data_dir=".simulations/")

    data = ss.simulate(
        t_eval=kwargs.pop("t_eval", np.linspace(0, 5e4, 1000)),
        init_state=init_state,
        num_units=kwargs.pop("num_units", 1e7),
        model_name="CRP3",
        report_polymers=kwargs.pop("report_polymers", False),
        report_sequences=kwargs.pop("report_sequences", False),
        **kwargs,
    )

    return data, ss.mdata



import numpy as np
from sparks.core import R_J


def arrhenius_eqn(A, Ea, R=R_J):
    def arrhenius(T_K):
        return A * np.exp(-Ea / (R * T_K))

    return arrhenius


def get_arrhenius_const(Ea_J_mol, k, T_K):
    return k / np.exp(-Ea_J_mol / (R_J * T_K))


def cal_to_J(cal):
    return cal * 4.184


def kcal_to_J(kcal):
    return kcal * 4184


def C_to_K(C):
    return C + 273.15


def K_to_C(K):
    return K - 273.15



from typing import Any, Callable, Dict, Generic, Optional, Tuple, Type, TypeAlias, cast
from functools import cached_property

import numpy as np
from scipy.integrate._ivp.ivp import OdeSolution

from sparks.core import NDArrayType
from sparks.models import StateType, ODEFunction, RateType

# SciPy RHS: f(t, y) -> ydot
ScipyODEFun: TypeAlias = Callable[[float, NDArrayType], NDArrayType]

# Sparks model: (dydt, k, t, s) -> dydt
ODEFun: TypeAlias = Callable[[StateType, RateType, float, StateType], StateType]

ODEStaticSolution: TypeAlias = Optional[Tuple[StateType, OdeSolution]]


# --------------------------------------------------------------------------------------
# Functions to compile a State class for fast evaluation in an ODE solver
# --------------------------------------------------------------------------------------


class _ViewBase:
    """
    Optimized read-only view of a State for fast ODE evaluation.
    """

    __slots__ = ("_y", "_t", "_memo")

    def __init__(self):
        self._y = None
        self._t = 0.0
        self._memo = {}

    @property
    def t(self) -> float:
        return self._t

    @t.setter
    def t(self, value: float) -> None:
        self._t = float(value)

    def __setattr__(self, name: str, value: float) -> None:
        if name in _ViewBase.__slots__:
            return object.__setattr__(self, name, value)
        if name == "t":
            return type(self).t.__set__(self, float(value))
        raise AttributeError("Input state is read-only.")


class _WriterBase:
    """
    Optimized writable view of a State for fast ODE evaluation.
    """

    __slots__ = ("_buf", "_t")

    def __init__(self):
        self._buf: Optional[NDArrayType] = None
        self._t: float = 0.0

    @property
    def t(self) -> float:
        return self._t

    @t.setter
    def t(self, value: float) -> None:
        self._t = float(value)


"""Cache for state view and writer types"""
_VIEW_CACHE: Dict[Tuple[Type[StateType], bool], Type[_ViewBase]] = {}
_WRITER_CACHE: Dict[Type[StateType], Type[_WriterBase]] = {}


def compile_state_to_view(
    state_cls: Type[StateType], *, with_derived: bool
) -> Type[_ViewBase]:
    key = (state_cls, bool(with_derived))
    View = _VIEW_CACHE.get(key)

    if View is not None:
        return View

    names = state_cls.get_species_names()

    class View(_ViewBase):
        pass

    # Fast fixed-index properties
    for i, name in enumerate(names):

        def _make_state_var(ix=i, n=name):

            def get(self: _ViewBase):
                return self._y[ix]

            get.__name__ = f"get_{n}"
            return property(get)

        setattr(View, name, _make_state_var())

    # bind @derived getters (memoized per call)
    if with_derived:
        species_set = set(names)

        for attr_name, attr in vars(state_cls).items():

            if attr_name in species_set:
                continue

            if getattr(attr, "__is_derived__", False):

                if isinstance(attr, property):
                    fn = attr.fget
                elif isinstance(attr, cached_property):
                    fn = attr.func
                else:
                    fn = attr

                if attr_name in species_set:
                    raise ValueError(
                        f"Derived name '{attr_name}' conflicts with species."
                    )

                def _make_derived(_n=attr_name, _f=fn):

                    def get(self: _ViewBase):
                        m = self._memo
                        if _n in m:
                            return m[_n]
                        val = _f(self)
                        m[_n] = val
                        return val

                    get.__name__ = f"get_derived_{_n}"
                    return property(get)

                setattr(View, attr_name, _make_derived())

    def __getattr__(self, name: str):
        raise AttributeError(f"Unknown field '{name}'. Valid: {names}")

    View.__getattr__ = __getattr__
    View.__name__ = f"View_{state_cls.__name__}"
    _VIEW_CACHE[key] = View
    return View


def compile_state_to_writer(state_cls: Type[StateType]) -> Type[_WriterBase]:

    Writer = _WRITER_CACHE.get(state_cls)
    if Writer is not None:
        return Writer

    names = state_cls.get_species_names()

    class Writer(_WriterBase):
        pass

    for i, name in enumerate(names):

        def _make_state_var(ix=i, n=name):

            def get(self: _WriterBase) -> float:
                return self._buf[ix]

            def set(self: _WriterBase, value: float) -> None:
                self._buf[ix] = float(value)

            get.__name__ = f"get_{n}"
            set.__name__ = f"set_{n}"
            return property(get, set)

        setattr(Writer, name, _make_state_var())

    def __getattr__(self, name: str):
        raise AttributeError(f"Unknown field '{name}'. Valid: {names}")

    Writer.__getattr__ = __getattr__
    Writer.__name__ = f"Writer_{state_cls.__name__}"
    _WRITER_CACHE[state_cls] = Writer
    return Writer


def make_scipy_rhs(
    state_cls: Type[StateType],
    k: RateType,
    *,
    attach_derived: bool = True,
    debug: bool = False,
    **bound_kwargs: Any,
) -> Callable[[ODEFun], ScipyODEFun]:
    """Optimization of State"""

    names = state_cls.get_species_names()  # Single source of truth
    if "t" in names:
        raise ValueError("State class should not have 't' as a species name.")

    View = compile_state_to_view(state_cls, with_derived=attach_derived)
    Writer = compile_state_to_writer(state_cls)

    view = View()
    writer = Writer()
    buf: Optional[NDArrayType] = None

    # ----------------- Decorator that returns SciPy's f(t,y) -----------------
    def decorator(model_fun: ODEFun) -> ScipyODEFun:

        # Guard against double wrapping
        if getattr(model_fun, "__is_scipy_rhs__", False):
            raise TypeError(
                "make_scipy_rhs received a SciPy (t,y)->ydot function; "
                "pass a dydt-first model `(dydt, k, t, s)` instead."
            )

        def rhs(t: float, y: NDArrayType) -> NDArrayType:
            nonlocal buf

            # Allocate or resize output buffer as needed
            if (buf is None) or (buf.shape != y.shape) or (buf.dtype != y.dtype):
                buf = np.empty_like(y)

            # Wire the view/writer to current arrays and time
            view._y = y
            view.t = t
            view._memo.clear()  # Fresh deriveds each call

            writer._buf = buf
            writer.t = t

            if debug:
                # Help catch missing assignments quickly
                buf[...] = np.nan
            else:
                buf.fill(0.0)

            # Give IDEs/types a hint; at runtime these are view/writer
            s = cast(StateType, view)
            dydt = cast(StateType, writer)

            extra = {}
            for key, val in bound_kwargs.items():
                if getattr(val, "_is_state_provider", False):
                    extra[key] = val(t)
                else:
                    extra[key] = val

            # Run the user's equations
            model_fun(dydt, k, t, s, **extra)

            # Ensure all outputs were written this call
            if debug and np.isnan(buf).any():
                missing = [names[i] for i, v in enumerate(buf) if np.isnan(v)]
                raise RuntimeError(f"RHS did not set all outputs: {missing}")

            return buf

        rhs.__is_scipy_rhs__ = True
        rhs.__wrapped_model__ = model_fun
        return rhs

    return decorator


def compile_odes(
    state_cls: Type[StateType],
    ode_fun: ODEFunction,
    k: RateType,
    # static_state: Optional[OdeSolution] = None,
    **kwargs: Any,
) -> ScipyODEFun:

    return make_scipy_rhs(state_cls, k, **kwargs)(ode_fun)


class OdeSolutionState(Generic[StateType]):
    """
    Wrap an OdeSolution and present it as a read-only State view at time t.
    """

    __slots__ = ("_sol", "_view", "_state_cls", "_debug")
    _is_state_provider = True

    def __init__(
        self,
        state_cls: Type[StateType],
        ode_sol: OdeSolution,
        *,
        attach_derived: bool = True,
        debug: bool = False,
    ):
        self._state_cls = state_cls
        self._sol = ode_sol

        View = compile_state_to_view(state_cls, with_derived=attach_derived)
        self._view = View()
        self._debug = debug

    def __call__(self, t: float) -> StateType:
        y = np.asarray(self._sol(t), dtype=float)

        # SciPy returns (n,) for scalar t, or (n, m) for vector t – handle scalar only here
        if y.ndim == 2:
            if y.shape[1] == 1:
                y = y[:, 0]
            else:
                raise ValueError(
                    "OdeSolutionState: OdeSolution returned multiple time points; expected scalar t."
                )

        if self._debug:
            expected = len(self._state_cls.get_species_names())
            if y.ndim != 1 or y.shape[0] != expected:
                raise ValueError(
                    f"OdeSolutionState: expected length {expected} for {self._state_cls.__name__}, got {y.shape}."
                )

        v = self._view
        v._y = y
        v.t = t
        v._memo.clear()
        return cast(StateType, v)



from enum import Enum
from dataclasses import dataclass

from sparks.utils.formulas import arrhenius_eqn, kcal_to_J, C_to_K, get_arrhenius_const
from sparks.core import R_cal

# +-----------------------------------------------------------
# | Monomer Rate Constants and Properties sourced from WATPOLY
# | (https://doi.org/10.1080/10601325.2017.1312678)
# | and other literature sources as noted below
# +-----------------------------------------------------------


class MonomerType(Enum):
    General = "General"
    AMS = "AMS"
    MMA = "MMA"
    BA = "BA"
    Sty = "Sty"
    AN = "AN"
    HEA = "HEA"
    BMA = "BMA"
    AA = "AA"
    EA = "EA"


@dataclass
class MonomerData:

    name: str
    id: MonomerType
    # id: str

    # Molecular weight of the monomer (g/mol)
    MW: float

    # Glass transition temperature of the monomer (K)
    Tg_m: float

    # Glass transition temperature of the polymer (K)
    Tg_p: float

    # Heat capacity of the monomer (cal/kg/K)
    Cp_m: float

    # Heat capacity of the polymer (cal/kg/K)
    Cp_p: float

    # Heat of reaction (cal/mol)
    deltaH: float

    # Density of monomer (g/L)
    dens_m: float

    # Density of polymer (g/L)
    dens_p: float

    # Rate of propagation (L/mol/min)
    kp: float

    # Rate of termination (L/mol/min)
    kt: float

    # Disproportionation to combination ratio
    ktd_ratio: float

    # Transfer to monomer rate (L/mol/min)
    kf_m: float

    # Transfer to polymer rate (L/mol/min)
    kf_p: float

    # Internal double bond rate of propagation (L/mol/min)
    kp_in: float

    # Terminal double bond rate of propagation (L/mol/min)
    kp_te: float

    # Reaction radius for segmental diffusion (L/g)
    delta: float

    # Critical free volume (L)
    Vf_c: float

    # Free volume of the monomer (L)
    Vf_m: float

    # Thermal expansion coefficient of the monomer (L/K)
    alpha_m: float

    # Free volume of the polymer (L)
    Vf_p: float

    # Thermal expansion coefficient of the polymer (L/K)
    alpha_p: float

    # Rate of decrease of kp
    B: float

    # Gel-effect model parameter
    m: float
    n: float

    # Rate of decrease of kt
    A: float

    # Onset pt. of translational diffusion-control
    K3: float

    # Avg. number of monomer units per chain
    n_s: float

    # Length of monomer unit per chain (cm)
    l0: float

    # Thermal (/self) initiation rate (L^2/mol^2/min)
    kth: float


# S. I. Cheong & A. Penlidis (2003)
# Modeling of the Copolymerization, with Depropagation, of α-Methyl Styrene
# and Methyl Methacrylate at an Elevated Temperature
# Table 1
kp_AMS_100C = 390 * 60  # L/mol/min
Ea_kp_AMS = kcal_to_J(13)  # J/mol
A_kp_AMS = get_arrhenius_const(Ea_kp_AMS, kp_AMS_100C, C_to_K(100))
kp_AMS = arrhenius_eqn(A_kp_AMS, Ea_kp_AMS)

kt_AMS_100C = 8.3e11  # L/mol/min
Ea_kt_AMS = kcal_to_J(2)
A_kt_AMS = get_arrhenius_const(Ea_kt_AMS, kt_AMS_100C, C_to_K(100))
kt_AMS = arrhenius_eqn(A_kt_AMS, Ea_kt_AMS)


def alpha_methyl_styrene(T_K: float) -> MonomerData:
    return MonomerData(
        name="alpha-methyl styrene",
        id=MonomerType.AMS,
        MW=118.16,
        Tg_m=150.15,
        Tg_p=449.15,
        Cp_m=400,
        Cp_p=400,
        deltaH=-1.7e4,
        dens_m=(0.875 - 0.000918 * (T_K - 273.15)) * 1000,
        dens_p=(1.15 - 0.000918 * (T_K - 273.15)) * 1000,
        kp=kp_AMS(T_K),
        kt=kt_AMS(T_K),
        ktd_ratio=0.07,
        kf_m=arrhenius_eqn(3.3615e9, 15177, R=R_cal)(T_K),
        kf_p=0.0,
        kp_in=0.0,
        kp_te=0.0,
        delta=0.001,
        Vf_c=arrhenius_eqn(1.2, 2220, R=R_cal)(T_K),
        Vf_m=0.025,
        alpha_m=0.001,
        Vf_p=0.025,
        alpha_p=0.00048,
        B=0.5,
        m=0.5,
        n=1.75,
        A=0.55,
        K3=1e10,
        n_s=120,
        l0=5.0e-8,
        kth=2.5e-9 * 60,  # Cheong 2003
    )


# kp/kt from Nising P., & Meyer T. Ind. Eng. Chem. Res., Vol. 43, No. 23, 2004
# - Ref: Brandup & Immergut, Polymer Handbook, 4th ed. (II-84)
# - - Ref: H. K. Mahabadi, K. F. O'Driscoll, J. Macromol. Sci. A: Chem., 11, 967 (1977).
def methyl_methacrylate(T_K: float) -> MonomerData:
    return MonomerData(
        name="methyl methacrylate",
        id=MonomerType.MMA,
        MW=100.12,
        Tg_m=167.1,
        Tg_p=378,
        Cp_m=411.1,
        Cp_p=400,
        deltaH=-1.381e4,
        dens_m=(0.9665 - 0.001164 * (T_K - 273.15)) * 1e3,
        dens_p=(1.195 - 0.00033 * (T_K - 273.15)) * 1e3,
        kp=arrhenius_eqn(2.952e7, 4353, R=R_cal)(T_K),
        kt=arrhenius_eqn(5.88e9, 701, R=R_cal)(T_K),
        ktd_ratio=arrhenius_eqn(1.6093, -440.12, R=R_cal)(T_K),
        kf_m=arrhenius_eqn(9.3435e4, 7475, R=R_cal)(T_K),
        kf_p=0.0,
        kp_in=0.0,
        kp_te=0.0,
        delta=0.001,
        Vf_c=arrhenius_eqn(0.7408, 1589.6, R=R_cal)(T_K),
        Vf_m=0.025,
        alpha_m=0.001,
        Vf_p=0.025,
        alpha_p=0.00048,
        B=1.0,
        m=0.5,
        n=1.75,
        A=1.11,
        K3=arrhenius_eqn(0.563, -8900, R=R_cal)(T_K),
        n_s=47,
        l0=6.9e-8,
        kth=arrhenius_eqn(2.26e-6, 6578, R=R_cal)(T_K),
    )



from __future__ import annotations
from dataclasses import dataclass

import numpy as np

from sparks.core import NDArrayType
from .model import RateCoefficients, ModelState, SequenceModelState, ChainModelState


# Needed for DAE
from scipy.optimize import root, least_squares


@dataclass
class Context:

    prev_solution = [0.5] * 6


def kinetic_odes(
    dydt: ModelState,
    k: RateCoefficients,
    t: NDArrayType,
    s: ModelState,
    context: Context = Context(),
) -> ModelState:
    """Defines the ODEs for the full FRP2 model."""

    dydt.t = t

    eqns = DAE_equations(s, k)
    result = least_squares(eqns, context.prev_solution, bounds=[[0] * 6, [1] * 6])

    if result.success:

        pA, pB, pAA, pBA, pAB, pBB = result.x
        context.prev_solution = result.x
        # print(f"Root found at t={t}: {result.x}")
    else:
        print(f"Root finding failed at t={t}: {result.message}")

        pA, pB, pAA, pBA, pAB, pBB = context.prev_solution

    PA = pA * s.P
    PB = pB * s.P
    PAA = pAA * pA * s.P
    PAB = pAB * pB * s.P
    PBA = pBA * pA * s.P
    PBB = pBB * pB * s.P

    kp = (
        # i=A, j=A
        (k.kpAA * pA * s.fA - k.kdAA * pAA * pA / s.M)
        # i=A, j=B
        + (k.kpAB * pA * s.fB - k.kdAB * pAB * pB / s.M)
        # i=B, j=A
        + (k.kpBA * pB * s.fA - k.kdBA * pBA * pA / s.M)
        # i=B, j=B
        + (k.kpBB * pB * s.fB - k.kdBB * pBB * pB / s.M)
    )

    kt = k.ktAA * pA * pA + k.ktBB * pB * pB + 2 * k.ktAB * pA * pB

    """ Concentration of Initiator """
    dydt.I = -k.kd * s.I

    """ Concentration of Free Radical, R """
    dydt.R = (
        # Formation by initiation
        +2 * k.f * k.kd * s.I
        # Consumption by propagation
        - k.kpAA * s.R * s.A
        - k.kpBB * s.R * s.B
    )

    """ Concentration of A Monomer """
    dydt.A = (
        # Consumption by propagation
        -k.kpAA * s.R * s.A
        - k.kpAA * PA * s.A
        - k.kpBA * PB * s.A
        # Formation by depropagation
        + k.kdAA * PAA
        + k.kdBA * PBA
    )

    """ Concentration of B Monomer """
    dydt.B = (
        # Consumption by propagation
        -k.kpBB * s.R * s.B
        - k.kpAB * PA * s.B
        - k.kpBB * PB * s.B
        # Formation by depropagation
        + k.kdAB * PAB
        + k.kdBB * PBB
    )

    """ Concentration of Polymer """
    dydt.P = k.kpAA * s.R * s.A + k.kpBB * s.R * s.B - kt * s.P**2

    return dydt


def chain_odes(
    dydt: ChainModelState,
    k: RateCoefficients,
    t: NDArrayType,
    cms: ChainModelState,
    ode_sol: ChainModelState,
) -> ChainModelState:

    dydt.t = t

    s = ModelState.from_array(ode_sol(t))

    ###########################################
    ### Calculate the pseudo rate constants ###
    ###########################################

    # Pseudo rate constant for propagation
    # Mistake in Eq. 52? kdij * Pij * pi / M
    # Should be kdij * Pij * pj / M
    kp = (
        # i=A, j=A
        (k.kpAA * s.pA * s.fA - k.kdAA * s.pAA * s.pA / s.M)
        # i=A, j=B
        + (k.kpAB * s.pA * s.fB - k.kdAB * s.pAB * s.pB / s.M)
        # i=B, j=A
        + (k.kpBA * s.pB * s.fA - k.kdBA * s.pBA * s.pA / s.M)
        # i=B, j=B
        + (k.kpBB * s.pB * s.fB - k.kdBB * s.pBB * s.pB / s.M)
    )

    kt = k.ktAA * s.pA * s.pA + k.ktBB * s.pB * s.pB + 2 * k.ktAB * s.pA * s.pB
    ktc = k.ktcAA * s.pA * s.pA + k.ktcBB * s.pB * s.pB + 2 * k.ktcAB * s.pA * s.pB
    ktd = k.ktdAA * s.pA * s.pA + k.ktdBB * s.pB * s.pB + 2 * k.ktdAB * s.pA * s.pB

    # Formation of polymer from active radicals
    poly_form = k.kpAA * s.R * s.A + k.kpBB * s.R * s.B

    """ Active radicals (0th order moment) """
    dydt.lam_0 = (
        # Formation by propagation
        poly_form
        # Consumption by termination
        - kt * cms.lam_0**2
    )

    """ Active radicals (1st order moment) """
    dydt.lam_1 = (
        # Formation by propagation
        poly_form
        # Formation by propagation
        + kp * s.M * cms.lam_0
        # Consumption by termination
        - kt * cms.lam_0 * cms.lam_1
    )

    """Active radicals (2nd order moment)"""
    dydt.lam_2 = (
        # Formation by propagation
        poly_form
        # Formation by propagation
        + kp * s.M * (cms.lam_0 + 2 * cms.lam_1)
        # Consumption by termination
        - kt * cms.lam_0 * cms.lam_2
    )

    """Dead radicals (0th order moment)"""
    dydt.mu_0 = (
        # Formation by termination
        +(0.5 * ktc + ktd)
        * cms.lam_0**2
    )

    """Dead radicals (1st order moment)"""
    dydt.mu_1 = (
        # Formation by termination
        +(ktc + ktd)
        * cms.lam_0
        * cms.lam_1
    )

    """Dead radicals (2nd order moment)"""
    dydt.mu_2 = (
        # Formation by termination (comb.)
        +ktc * (cms.lam_0 * cms.lam_2 + cms.lam_1**2)
        # Formation by termination (disp.)
        + ktd * cms.lam_0 * cms.lam_2
    )

    return dydt


def sequence_odes(
    dydt: SequenceModelState,
    k: RateCoefficients,
    t: NDArrayType,
    sms: SequenceModelState,
    ode_sol: SequenceModelState,
) -> SequenceModelState:

    dydt.t = t

    s = ModelState.from_array(ode_sol(t))

    sms.aSA0 = sms.aSA0
    sms.aSB0 = sms.aSB0
    sms.aSA1 = sms.aSA1
    sms.aSB1 = sms.aSB1
    sms.aSA2 = sms.aSA2
    sms.aSB2 = sms.aSB2

    # Rates of termination
    # Ctrij_x = ktij * sms.aSijx * sms.aSij0
    RtAA_0 = k.ktAA * sms.aSA0 * sms.aSA0
    RtAB_0 = k.ktAB * sms.aSA0 * sms.aSB0
    RtBA_0 = k.ktAB * sms.aSB0 * sms.aSA0
    RtBB_0 = k.ktBB * sms.aSB0 * sms.aSB0
    RtAA_1 = k.ktAA * sms.aSA1 * sms.aSA0
    RtAB_1 = k.ktAB * sms.aSA1 * sms.aSB0
    RtBA_1 = k.ktAB * sms.aSB1 * sms.aSA0
    RtBB_1 = k.ktBB * sms.aSB1 * sms.aSB0
    RtAA_2 = k.ktAA * sms.aSA2 * sms.aSA0
    RtAB_2 = k.ktAB * sms.aSA2 * sms.aSB0
    RtBA_2 = k.ktAB * sms.aSB2 * sms.aSA0
    RtBB_2 = k.ktBB * sms.aSB2 * sms.aSB0

    a10_frac = sms.aSA1 / sms.aSA0 if sms.aSA0 > 0 else 0
    b10_frac = sms.aSB1 / sms.aSB0 if sms.aSB0 > 0 else 0
    a20_frac = sms.aSA2 / sms.aSA0 if sms.aSA0 > 0 else 0
    b20_frac = sms.aSB2 / sms.aSB0 if sms.aSB0 > 0 else 0

    """Active A sequences (0th order moment)"""
    dydt.aSA0 = (
        # Formation by propagation (PA + A -> PA, PB + A -> PA)
        +k.kpAA * s.A * (s.R + sms.aSA0)
        + k.kpBA * s.A * (s.RB + sms.aSB0)
        # Consumption by propagation (PA + A -> PA, PA + B -> PB)
        - k.kpAA * s.A * sms.aSA0
        - k.kpAB * s.B * (s.RA + sms.aSA0)
        # Formation by depropagation (PAA -> PA + A, PAB -> PA + B)
        + k.kdAA * s.pAA * sms.aSA0
        + k.kdAB * s.pAB * sms.aSB0
        # Consumption by depropagation (PAA -> PA + A, PBA -> PB + A)
        - k.kdAA * s.pAA * sms.aSA0
        - k.kdBA * s.pBA * sms.aSA0
        # Consumption by termination (PA + PA -> D, PA + PB -> D)
        - RtAA_0
        - RtAB_0
    )

    """Active B sequences (0th order moment)"""
    dydt.aSB0 = (
        # Formation by propagation (PB + B -> PB, PA + B -> PB)
        +k.kpBB * s.B * (s.R + sms.aSB0)
        + k.kpAB * s.B * (s.RA + sms.aSA0)
        # Consumption by propagation (PB + B -> PB, PB + A -> PA)
        - k.kpBB * s.B * sms.aSB0
        - k.kpBA * s.A * (s.RB + sms.aSB0)
        # Formation by depropagation (PBB -> PB + B, PBA -> PB + A)
        + k.kdBB * s.pBB * sms.aSB0
        + k.kdBA * s.pBA * sms.aSA0
        # Consumption by depropagation (PBB -> PB + B, PAB -> PA + B)
        - k.kdBB * s.pBB * sms.aSB0
        - k.kdAB * s.pAB * sms.aSB0
        # Consumption by termination (PB + PB -> D, PB + PA -> D)
        - RtBB_0
        - RtBA_0
    )

    """Active A sequences (1st order moment)"""
    dydt.aSA1 = (
        # Formation by propagation (PA + A -> PA, PB + A -> PA)
        +k.kpAA * s.A * (s.R + sms.aSA0 + sms.aSA1)
        + k.kpBA * s.A * (s.RB + sms.aSB0)
        # Consumption by propagation (PA + A -> PA, PA + B -> PB)
        - k.kpAA * s.A * sms.aSA1
        - k.kpAB * s.B * (s.RA + sms.aSA1)
        # Formation by depropagation (PAA -> PA + A, PAB -> PA + B)
        + k.kdAA * s.pAA * (sms.aSA1 - sms.aSA0)
        + k.kdAB * s.pAB * a10_frac * sms.aSB0
        # Consumption by depropagation (PAA -> PA + A, PBA -> PB + A)
        - k.kdAA * s.pAA * sms.aSA1
        - k.kdBA * s.pBA * sms.aSA0
        # Consumption by termination (PA + PA -> D, PA + PB -> D)
        - RtAA_1
        - RtAB_1
    )

    """Active B sequences (1st order moment)"""
    dydt.aSB1 = (
        # Formation by propagation (PB + B -> PB, PA + B -> PB)
        +k.kpBB * s.B * (s.R + sms.aSB0 + sms.aSB1)
        + k.kpAB * s.B * (s.RA + sms.aSA0)
        # Consumption by propagation (PB + B -> PB, PB + A -> PA)
        - k.kpBB * s.B * sms.aSB1
        - k.kpBA * s.A * (s.RB + sms.aSB1)
        # Formation by depropagation (PBB -> PB + B, PBA -> PB + A)
        + k.kdBB * s.pBB * (sms.aSB1 - sms.aSB0)
        + k.kdBA * s.pBA * b10_frac * sms.aSA0
        # Consumption by depropagation (PBB -> PB + B, PAB -> PA + B)
        - k.kdBB * s.pBB * sms.aSB1
        - k.kdAB * s.pAB * sms.aSB0
        # Consumption by termination (PB + PB -> D, PB + PA -> D)
        - RtBB_1
        - RtBA_1
    )

    """Active A sequences (2nd order moment)"""
    dydt.aSA2 = (
        # Formation by propagation (PA + A -> PA, PB + A -> PA)
        +k.kpAA * s.A * (s.R + sms.aSA0 + 2 * sms.aSA1 + sms.aSA2)
        + k.kpBA * s.A * (s.RB + sms.aSB0)
        # Consumption by propagation (PA + A -> PA, PA + B -> PB)
        - k.kpAA * s.A * sms.aSA2
        - k.kpAB * s.B * (s.RA + sms.aSA2)
        # Formation by depropagation (PAA -> PA + A, PAB -> PA + B)
        + k.kdAA * s.pAA * (sms.aSA2 - 2 * sms.aSA1 + sms.aSA0)
        + k.kdAB * s.pAB * a20_frac * sms.aSB0
        # Consumption by depropagation (PAA -> PA + A, PBA -> PB + A)
        - k.kdAA * s.pAA * sms.aSA2
        - k.kdBA * s.pBA * sms.aSA0
        # Consumption by termination (PA + PA -> D, PA + PB -> D)
        - RtAA_2
        - RtAB_2
    )

    """Active B sequences (2nd order moment)"""
    dydt.aSB2 = (
        # Formation by propagation (PB + B -> PB, PA + B -> PB)
        +k.kpBB * s.B * (s.R + sms.aSB0 + 2 * sms.aSB1 + sms.aSB2)
        + k.kpAB * s.B * (s.RA + sms.aSA0)
        # Consumption by propagation (PB + B -> PB, PB + A -> PA)
        - k.kpBB * s.B * sms.aSB2
        - k.kpBA * s.A * (s.RB + sms.aSB2)
        # Formation by depropagation (PBB -> PB + B, PBA -> PB + A)
        + k.kdBB * s.pBB * (sms.aSB2 - 2 * sms.aSB1 + sms.aSB0)
        + k.kdBA * s.pBA * b20_frac * sms.aSA0
        # Consumption by depropagation (PBB -> PB + B, PAB -> PA + B)
        - k.kdBB * s.pBB * sms.aSB2
        - k.kdAB * s.pAB * sms.aSB0
        # Consumption by termination (PB + PB -> D, PB + PA -> D)
        - RtBB_2
        - RtBA_2
    )

    """Inactive A sequences (0th order moment)"""
    dydt.iSA0 = (
        # Formation by cross-propagation (PA + B -> PB)
        +k.kpAB * s.B * (s.RA + sms.aSA0)
        # Consumption by depropagation (PAB -> PA + B)
        - k.kdAB * s.pAB * sms.aSB0
        # Formation by termination (PA + PA -> D, PA + PB -> D)
        + RtAA_0
        + RtAB_0
    )

    """Inactive B sequences (0th order moment)"""
    dydt.iSB0 = (
        # Formation by cross-propagation (PB + A -> PA)
        +k.kpBA * s.A * (s.RB + sms.aSB0)
        # Consumption by depropagation (PBA -> PB + A)
        - k.kdBA * s.pBA * sms.aSA0
        # Formation by termination (PB + PB -> D, PB + PA -> D)
        + RtBB_0
        + RtBA_0
    )

    """Inactive A sequences (1st order moment)"""
    dydt.iSA1 = (
        # Formation by cross-propagation (PA + B -> PB)
        +k.kpAB * s.B * (s.RA + sms.aSA1)
        # Consumption by depropagation (PAB -> PA + B)
        - k.kdAB * s.pAB * a10_frac * sms.aSB0
        # Formation by termination (PA + PA -> D, PA + PB -> D)
        + RtAA_1
        + RtAB_1
    )

    """Inactive B sequences (1st order moment)"""
    dydt.iSB1 = (
        # Formation by cross-propagation (PB + A -> PA)
        +k.kpBA * s.A * (s.RB + sms.aSB1)
        # Consumption by depropagation (PBA -> PB + A)
        - k.kdBA * s.pBA * b10_frac * sms.aSA0
        # Formation by termination (PB + PB -> D, PB + PA -> D)
        + RtBB_1
        + RtBA_1
    )

    """Inactive A sequences (2nd order moment)"""
    dydt.iSA2 = (
        # Formation by cross-propagation (PA + B -> PB)
        +k.kpAB * s.B * (s.RA + sms.aSA2)
        # Consumption by depropagation (PAB -> PA + B)
        - k.kdAB * s.pAB * a20_frac * sms.aSB0
        # Formation by termination (PA + PA -> D, PA + PB -> D)
        + RtAA_2
        + RtAB_2
    )

    """Inactive B sequences (2nd order moment)"""
    dydt.iSB2 = (
        # Formation by cross-propagation (PB + A -> PA)
        +k.kpBA * s.A * (s.RB + sms.aSB2)
        # Consumption by depropagation (PBA -> PB + A)
        - k.kdBA * s.pBA * b20_frac * sms.aSA0
        # Formation by termination (PB + PB -> D, PB + PA -> D)
        + RtBB_2
        + RtBA_2
    )

    return dydt


def DAE_equations(s: ModelState, k: RateCoefficients):

    def eqns(x):

        _pA, _pB, _pAA, _pBA, _pAB, _pBB = x

        _PA = _pA * s.P
        _PB = _pB * s.P
        _PAA = _pAA * _pA * s.P
        _PAB = _pAB * _pB * s.P
        _PBA = _pBA * _pA * s.P
        _PBB = _pBB * _pB * s.P

        eqs = np.zeros(6)

        # dPA/dt = 0
        eqs[0] = k.kpBA * _PB * s.A + k.kdAB * _PAB - k.kpAB * _PA * s.B - k.kdBA * _PBA

        # dPAA/dt = 0
        eqs[1] = (
            # Formation by propagation
            k.kpAA * _PA * s.A
            # Consumption by propagation
            - _PAA * (k.kpAA * s.A + k.kpAB * s.B)
            # Formation by depropagation
            + k.kdAA * _pAA * _PAA
            + k.kdAB * _pAA * _PAB
            # Consumption by depropagation
            - k.kdAA * _PAA
        )

        # dPBB/dt = 0
        eqs[2] = (
            # Formation by propagation
            k.kpBB * _PB * s.B
            # Consumption by propagation
            - _PBB * (k.kpBB * s.B + k.kpBA * s.A)
            # Formation by depropagation
            + k.kdBB * _pBB * _PBB
            + k.kdBA * _pBB * _PBA
            # Consumption by depropagation
            - k.kdBB * _PBB
        )

        # Identity equations
        eqs[3] = 1 - _pA - _pB
        eqs[4] = 1 - _pAA - _pBA
        eqs[5] = 1 - _pAB - _pBB

        return eqs

    return eqns



from ..model import ModelState, RateCoefficients


def k_FRP2_from_ratios(
    rA: float,
    rB: float,
    rX: float = 1.0,
    KAA: float = 0.0,
    KAB: float = 0.0,
    KBA: float = 0.0,
    KBB: float = 0.0,
    kpAA: float = 1.0,
    kd: float = 3.0e-06,
    kt: float = 1.0e7,
    ktd_frac: float = 1.0,
) -> RateCoefficients:

    # rX = kpAA / kpBB
    kpBB = kpAA / rX

    kpAB = kpAA / rA
    kpBA = kpBB / rB

    kdAA = KAA * kpAA
    kdAB = KAB * kpAB
    kdBA = KBA * kpBA
    kdBB = KBB * kpBB

    ktdAA = ktd_frac * kt
    ktdAB = ktd_frac * kt
    ktdBB = ktd_frac * kt
    ktcAA = (1 - ktd_frac) * kt
    ktcAB = (1 - ktd_frac) * kt
    ktcBB = (1 - ktd_frac) * kt

    return RateCoefficients(
        kd=kd,
        f=0.5,
        kpAA=kpAA,
        kpAB=kpAB,
        kpBA=kpBA,
        kpBB=kpBB,
        kdAA=kdAA,
        kdAB=kdAB,
        kdBA=kdBA,
        kdBB=kdBB,
        ktcAA=ktcAA,
        ktcAB=ktcAB,
        ktcBB=ktcBB,
        ktdAA=ktdAA,
        ktdAB=ktdAB,
        ktdBB=ktdBB,
    )


def k_FRP2_equal_from_ratio(
    rA: float,
    rB: float,
    K: float,
    **kwargs,
) -> RateCoefficients:

    return k_FRP2_from_ratios(
        rA=rA,
        rB=rB,
        KAA=K,
        KAB=K,
        KBA=K,
        KBB=K,
        **kwargs,
    )


def c0_FRP2(
    fA0: float = 0.50, M0: float = 1.0, M0_I0_ratio: float = 1000.0
) -> ModelState:

    A0 = fA0 * M0
    B0 = (1 - fA0) * M0
    I0 = M0 / M0_I0_ratio

    return ModelState(I=I0, A=A0, B=B0)



from abc import ABC, abstractmethod

import numpy as np

from sparks.models import BaseModelProtocol, RateType, StateType, SimulationData


class Simulator(ABC):
    """Base class for polymerization simulators"""

    def __init__(self, model: BaseModelProtocol, k: RateType):
        self.model = model
        self.k = k

    @abstractmethod
    def simulate(
        self, t_eval: np.ndarray, init_state: StateType, **kwargs
    ) -> SimulationData[StateType]:
        """Run simulation and return time points and final state"""
        pass



from typing import TypeAlias
import numpy as np
from numpy.typing import NDArray


NDArrayType: TypeAlias = NDArray[np.floating]


#################
### Constants ###
#################

# Small value to avoid division by zero
EPS = 1e-40

# Avogadro's number (mol^-1)
NA = 6.022e23

# Gas constants
R_J = 8.314  # J/mol/K
R_cal = 1.9872036  # cal/mol/K
